import ast
import html
import inspect
import json
import os
import traceback
from collections import defaultdict, Sequence, Set, Mapping, deque, namedtuple
from datetime import datetime
from functools import partial
from itertools import chain, islice
from threading import Lock

from asttokens import ASTTokens

from birdseye.app import db, Function, Call
from birdseye.cheap_repr import cheap_repr
from birdseye.tracer import TreeTracerBase, loops, TracedFile
from birdseye.utils import safe_qualname, correct_type, exception_string

CodeInfo = namedtuple('CodeInfo', 'db_func traced_file')


class BirdsEye(TreeTracerBase):
    def __init__(self):
        super().__init__()
        self._code_infos = {}

    def parse_extra(self, root, source, filename):
        for node in ast.walk(root):
            node._loops = loops(node)

    def compile(self, source, filename):
        traced_file = super(BirdsEye, self).compile(source, filename)
        traced_file.tokens = ASTTokens(source, tree=traced_file.root)
        return traced_file

    def before_stmt(self, node, frame):
        if frame.f_code not in self._code_infos:
            return
        if isinstance(node.parent, (ast.For, ast.While)) and node is node.parent.body[0]:
            iteration = self.stack[frame].iteration
            for i, loop_node in enumerate(node._loops):
                loop = iteration.loops[loop_node._tree_index]
                if i == len(node._loops) - 1:
                    loop.append(Iteration())
                else:
                    iteration = loop.last()

    def after_expr(self, node, frame, value):
        if frame.f_code not in self._code_infos:
            return
        if is_obvious_builtin(node, self.stack[frame]):
            return

        self._set_node_value(
            node, frame,
            expand(value, level=max(1, 3 - len(node._loops))))

    def _set_node_value(self, node, frame, value):
        iteration = self.stack[frame].iteration
        for i, loop_node in enumerate(node._loops):
            loop = iteration.loops[loop_node._tree_index]
            iteration = loop.last()
        iteration.vals[node._tree_index] = value

    def after_stmt(self, node, frame, exc_value, exc_traceback):
        if frame.f_code not in self._code_infos:
            return
        expression_stack = self.stack[frame].expression_stack
        if expression_stack:
            self._set_node_value(
                expression_stack[-1], frame,
                [exception_string(exc_value), -1])

    def enter_call(self, enter_info):
        frame = enter_info.current_frame
        if frame.f_code not in self._code_infos:
            return
        frame_info = self.stack[frame]
        frame_info.start_time = datetime.now()
        frame_info.iteration = Iteration()
        arg_info = inspect.getargvalues(frame)
        arg_names = chain(arg_info[0], arg_info[1:3])
        f_locals = arg_info[3].copy()
        arguments = [(name, f_locals.pop(name)) for name in arg_names if name] + list(f_locals.items())
        frame_info.arguments = json.dumps([[k, cheap_repr(v)] for k, v in arguments])

    def exit_call(self, exit_info):
        frame = exit_info.current_frame
        if frame.f_code not in self._code_infos:
            return
        frame_info = self.stack[frame]
        top_iteration = frame_info.iteration

        loop_iterations = top_iteration.extract_iterations()['loops']

        expr_values = _deep_dict()

        def extract_values(iteration, path):
            for k, v in iteration.vals.items():
                full_path = (k,) + path
                d = expr_values
                for path_k in full_path[:-1]:
                    d = d[path_k]
                d[full_path[-1]] = v

            for loop in iteration.loops.values():
                for i, iteration in enumerate(loop):
                    extract_values(iteration, path + (i,))

        extract_values(top_iteration, ())

        db_func = self._code_infos[frame.f_code].db_func
        exc = exit_info.exc_value
        if exc:
            traceback_str = ''.join(traceback.format_exception(type(exc), exc, exit_info.exc_tb))
            exception = exception_string(exc)
        else:
            traceback_str = exception = None

        call = Call(function=db_func,
                    arguments=frame_info.arguments,
                    return_value=cheap_repr(exit_info.return_value),
                    exception=exception,
                    traceback=traceback_str,
                    data=json.dumps(dict(
                        expr_values=expr_values,
                        loop_iterations=loop_iterations,
                        type_names=type_registry.names(),
                        num_special_types=type_registry.num_special_types,
                    )),
                    start_time=frame_info.start_time)
        db.session.add(call)
        db.session.commit()

    def __call__(self, func):
        new_func = super(BirdsEye, self).__call__(func)
        code_info = self._code_infos.get(new_func.__code__)
        if code_info:
            return new_func
        lines, start_lineno = inspect.getsourcelines(func)
        end_lineno = start_lineno + len(lines)
        name = safe_qualname(func)
        filename = os.path.abspath(inspect.getsourcefile(func))

        traced_file = new_func.traced_file
        traced_file.root._depth = 0
        for node in ast.walk(traced_file.root):
            for child in ast.iter_child_nodes(node):
                child._depth = node._depth + 1

        positions = []
        node_loops = {}
        for node in traced_file.nodes:
            if isinstance(node, ast.expr):
                node_type = 'expr'
                if (isinstance(node, (ast.Num, ast.Str)) or
                        isinstance(getattr(node, 'ctx', None),
                                   (ast.Store, ast.Del))):
                    continue
            elif isinstance(node, (ast.While, ast.For)):
                node_type = 'loop'
            else:
                continue
            assert isinstance(node, ast.AST)
            if not start_lineno <= node.lineno <= end_lineno:
                continue
            if (isinstance(node, ast.UnaryOp) and
                    isinstance(node.op, (ast.UAdd, ast.USub)) and
                    isinstance(node.operand, ast.Num)):
                continue
            start, end = traced_file.tokens.get_text_range(node)
            if start == end == 0:
                continue
            positions.append((start, 1, node._depth,
                              '<span data-index="%s" data-type="%s">' % (node._tree_index, node_type)))
            positions.append((end, 0, node._depth, '</span>'))
            if node._loops:
                node_loops[node._tree_index] = [n._tree_index for n in node._loops]
        positions.append((len(traced_file.source), None, None, ''))
        positions.sort()

        html_body = []
        start = 0
        for pos, _, _, part in positions:
            html_body.append(html.escape(traced_file.source[start:pos]))
            html_body.append(part)
            start = pos
        html_body = ''.join(html_body)
        html_body = '\n'.join(html_body.split('\n')[start_lineno - 1:end_lineno - 1])

        db_args = dict(file=filename,
                       name=name,
                       html_body=html_body,
                       lineno=start_lineno,
                       data=json.dumps(
                           dict(
                               node_loops=node_loops,
                           ),
                           sort_keys=True,
                       ))
        db_func = Function.query.filter_by(**db_args).one_or_none()
        if not db_func:
            db_func = Function(**db_args)
            db.session.add(db_func)
            db.session.commit()
        self._code_infos[new_func.__code__] = CodeInfo(db_func, traced_file)

        return new_func


def _deep_dict():
    return defaultdict(_deep_dict)


class IterationList(object):
    side_len = 3

    def __init__(self):
        self.start = []
        self.end = deque(maxlen=self.side_len)
        self.length = 0

    def append(self, x):
        if self.length < self.side_len:
            self.start.append(x)
        else:
            self.end.append(x)
        x.index = self.length
        self.length += 1

    def __iter__(self):
        return chain(self.start, self.end)

    def last(self):
        if self.end:
            return self.end[-1]
        else:
            return self.start[-1]


class Iteration(object):
    def __init__(self):
        self.vals = {}
        self.loops = defaultdict(IterationList)
        self.index = None

    def extract_iterations(self):
        return {
            'index': self.index,
            'loops': {
                k: [it.extract_iterations() for it in v]
                for k, v in self.loops.items()
            }
        }


class TypeRegistry(object):
    def __init__(self):
        self.lock = Lock()
        self.data = defaultdict(lambda: len(self.data))
        basic_types = [type(None), bool, int, float, complex]  # TODO long, unicode
        special_types = basic_types + [list, dict, tuple, set, frozenset, str, bytes]
        self.num_basic_types = len(basic_types)
        self.num_special_types = len(special_types)
        for t in special_types:
            _ = self.data[t]

    def __getitem__(self, item):
        t = correct_type(item)
        with self.lock:
            return self.data[t]

    def names(self):
        rev = dict((v, k) for k, v in self.data.items())
        return [safe_qualname(rev[i]) for i in range(len(rev))]


type_registry = TypeRegistry()


def expand(val, level=3):
    result = []
    type_index = type_registry[val]
    result += [cheap_repr(val), type_index]
    if type_index < type_registry.num_basic_types or level == 0:
        return result
    exp = partial(expand, level=level - 1)

    # noinspection PyBroadException
    try:
        length = len(val)
    except:
        pass
    else:
        result += ['len() = %s' % length]

    if isinstance(val, (str, bytes, range)):  # TODO unicode, xrange
        return result
    if isinstance(val, Sequence):
        if len(val) <= 8:
            indices = range(len(val))
        else:
            indices = chain(range(3), range(len(val) - 3, len(val)))
        for i in indices:
            result += [(str(i), exp(val[i]))]
    elif isinstance(val, Mapping):
        for k, v in islice(val.items(), 10):  # TODO iteritems
            result += [(cheap_repr(k), exp(v))]
    elif isinstance(val, Set):
        for i, v in enumerate(islice(val, 6)):
            result += [('<%s>' % i, exp(v))]

    d = getattr(val, '__dict__', None)
    if d:
        for k, v in islice(d.items(), 50):  # TODO iteritems
            if isinstance(v, TracedFile):
                continue
            result += [(str(k), exp(v))]
    else:
        slots = getattr(val, '__slots__', None)
        if slots:
            for s in slots:
                try:
                    attr = getattr(val, s)
                except AttributeError:
                    pass
                else:
                    result += [(str(s), exp(attr))]
    return result


def is_obvious_builtin(node, frame_info):
    try:
        value = frame_info.expression_values[node]
        # noinspection PyUnresolvedReferences
        is_top_level_builtin = ((isinstance(node, ast.Name) and
                                 node.id in __builtins__ and
                                 __builtins__[node.id] is value) or
                                isinstance(node, ast.NameConstant))
        if is_top_level_builtin:
            return True
            # if not isinstance(node, ast.Attribute):
            #     return False
            # name = node.attr
            # if name != value.__name__:
            #     return False
            # if value.__self__ is not frame_info.expression_values[node.value]:
            #     return False
            # TODO complete for methods
    except AttributeError:
        # return False
        raise



import json
import os

from flask import Flask
from flask_humanize import Humanize
from flask_sqlalchemy import SQLAlchemy
from humanize import naturaltime
from markupsafe import Markup
from sqlalchemy import Sequence, UniqueConstraint

app = Flask('birdseye')
app.config.update(dict(
    SQLALCHEMY_DATABASE_URI=os.environ.get('BIRDSEYE_DB', 'sqlite:////tmp/test.db'),
    SQLALCHEMY_TRACK_MODIFICATIONS=False),
)
db = SQLAlchemy(app)
Humanize(app)


class Call(db.Model):
    id = db.Column(db.Integer, Sequence('call_id_seq'), primary_key=True)
    function_id = db.Column(db.Integer, db.ForeignKey('function.id'))
    function = db.relationship('Function', backref=db.backref('calls', lazy='dynamic'))
    arguments = db.Column(db.Text)
    return_value = db.Column(db.Text)
    exception = db.Column(db.Text)
    traceback = db.Column(db.Text)
    data = db.Column(db.Text)
    start_time = db.Column(db.DateTime)

    @property
    def pretty_start_time(self):
        return self._pretty_time(self.start_time)

    @staticmethod
    def _pretty_time(dt):
        if not dt:
            return ''
        return Markup('%s (%s)' % (
            dt.strftime('%Y-%m-%d&nbsp;%H:%M:%S'),
            naturaltime(dt)))

    @property
    def state_icon(self):
        return Markup('<span class="glyphicon glyphicon-%s" '
                      'style="color: %s"></span>' % (
                          ('ok', 'green') if self.success else
                          ('remove', 'red')))

    @property
    def success(self):
        if self.exception:
            assert self.traceback
            assert self.return_value == 'None'
            return False
        else:
            assert not self.traceback
            return True

    @property
    def result(self):
        if self.success:
            return self.return_value
        else:
            return self.exception

    @property
    def arguments_list(self):
        return json.loads(self.arguments)


class Function(db.Model):
    id = db.Column(db.Integer, Sequence('function_id_seq'), primary_key=True)
    file = db.Column(db.Text)
    name = db.Column(db.Text)
    html_body = db.Column(db.Text)
    lineno = db.Column(db.Integer)
    data = db.Column(db.Text)

    __table_args__ = (UniqueConstraint('file', 'name', 'html_body', 'lineno', 'data',
                                       name='everything_unique'),)


db.create_all()



import ntpath
import os
import traceback
import types

from littleutils import strip_required_prefix
from qualname import qualname


def path_leaf(path):
    # http://stackoverflow.com/a/8384788/2482744
    head, tail = ntpath.split(path)
    return tail or ntpath.basename(head)


def all_file_paths():
    from birdseye.app import db, Function
    return [f[0] for f in db.session.query(Function.file).distinct()]


def short_path(path):
    return strip_required_prefix(path, os.path.commonprefix(all_file_paths())) or path_leaf(path)


def safe_qualname(obj):
    result = safe_qualname._cache.get(obj)
    if not result:
        try:
            result = qualname(obj)
        except AttributeError:
            result = obj.__name__
        if '<locals>' not in result:
            safe_qualname._cache[obj] = result
    return result


safe_qualname._cache = {}


def correct_type(obj):
    # TODO handle case where __class__ has been assigned
    t = type(obj)
    if t is getattr(types, 'InstanceType', None):
        t = obj.__class__
    return t


def iter_get(it, n):
    n_original = n
    if n < 0:
        n = -n - 1
        it = reversed(it)
    else:
        it = iter(it)
    try:
        while n > 0:
            next(it)
            n -= 1
        return next(it)
    except StopIteration as e:
        raise IndexError(n_original) from e


def exception_string(exc):
    return ''.join(traceback.format_exception_only(type(exc), exc))



import ast
import functools
import inspect
import threading
from collections import namedtuple
from copy import deepcopy
from functools import partial

from littleutils import file_to_string


class TracedFile(object):
    def __init__(self, tracer, source, filename):
        self.tracer = tracer
        self.source = source
        self.filename = filename
        root = ast.parse(source, filename)

        def set_basic_node_attributes():
            self.nodes = []
            for node in ast.walk(root):
                for child in ast.iter_child_nodes(node):
                    child.parent = node
                node.traced_file = self
                node._tree_index = len(self.nodes)
                self.nodes.append(node)

        set_basic_node_attributes()

        new_root = self.tracer.parse_extra(root, source, filename)
        if new_root is not None:
            root = new_root

        set_basic_node_attributes()

        self.root = root

        root = deepcopy(root)
        root = _NodeVisitor().visit(root)

        self.code = compile(root, filename, "exec")


class FrameInfo(object):
    def __init__(self):
        self.statement = None
        self.expression_stack = []
        self.expression_values = {}
        self.return_node = None


EnterCallInfo = namedtuple('EnterCallInfo', 'call_node enter_node caller_frame current_frame')
ExitCallInfo = namedtuple('ExitCallInfo', 'call_node return_node caller_frame current_frame '
                                          'return_value exc_value exc_tb')


class TreeTracerBase(object):
    def __init__(self):
        self._thread_local = threading.local()

    @property
    def stack(self):
        result = getattr(self._thread_local, 'stack', None)
        if result is not None:
            return result
        result = self._thread_local.stack = {}
        return result

    @functools.lru_cache()
    def compile(self, source, filename):
        return TracedFile(self, source, filename)

    def exec_string(self, source, filename, globs=None, locs=None):
        traced_file = self.compile(source, filename)
        globs = globs or {}
        locs = locs or {}
        globs = dict(globs, **self._trace_methods_dict(traced_file))
        exec(traced_file.code, globals=globs, locals=locs)

    def _trace_methods_dict(self, traced_file):
        return {f.__name__: partial(f, traced_file)
                for f in [
                    self._treetrace_hidden_with_stmt,
                    self._treetrace_hidden_before_expr,
                    self._treetrace_hidden_after_expr,
                ]}

    def __call__(self, func):
        filename = inspect.getsourcefile(func)
        source = file_to_string(filename)
        traced_file = self.compile(source, filename)
        func.__globals__.update(self._trace_methods_dict(traced_file))

        code_options = []

        def find_code(root_code):
            for const in root_code.co_consts:
                if not inspect.iscode(const):
                    continue
                matches = (const.co_firstlineno == func.__code__.co_firstlineno and
                           const.co_name == func.__code__.co_name)
                if matches:
                    code_options.append(const)
                find_code(const)

        find_code(traced_file.code)
        if len(code_options) > 1:
            assert func.__code__.co_name == (lambda: 0).__code__.co_name
            raise ValueError("Failed to trace lambda. Convert the function to a def.")
        new_func_code = code_options[0]

        # http://stackoverflow.com/a/13503277/2482744
        # TODO python 2 methods
        new_func = type(func)(new_func_code, func.__globals__, func.__name__, func.__defaults__, func.__closure__)
        new_func = functools.update_wrapper(new_func, func)
        new_func.__kwdefaults__ = func.__kwdefaults__
        new_func.traced_file = traced_file
        return new_func

    def _treetrace_hidden_with_stmt(self, traced_file, _tree_index):
        node = traced_file.nodes[_tree_index]
        frame = inspect.currentframe().f_back
        return _StmtContext(self, node, frame)

    def _treetrace_hidden_before_expr(self, traced_file, _tree_index):
        node = traced_file.nodes[_tree_index]
        assert isinstance(node, ast.expr)
        frame = inspect.currentframe().f_back

        frame_info = self.stack.get(frame)
        if frame_info is None:
            return None

        frame_info.expression_stack.append(node)

        self.before_expr(node, frame)
        return node

    def _treetrace_hidden_after_expr(self, _, node, value):
        if node is None:
            return value
        frame = inspect.currentframe().f_back
        self.stack[frame].expression_stack.pop()
        self.stack[frame].expression_values[node] = value
        self.after_expr(node, frame, value)
        return value

    def _enter_call(self, enter_node, current_frame):
        caller_frame, call_node = self._get_caller_stuff(current_frame)
        self.stack[current_frame] = FrameInfo()
        self.enter_call(EnterCallInfo(call_node, enter_node, caller_frame, current_frame))

    def _get_caller_stuff(self, frame):
        caller_frame = frame.f_back
        call_node = None
        if caller_frame in self.stack:
            expression_stack = self.stack[caller_frame].expression_stack
            if expression_stack:
                call_node = expression_stack[-1]
        return caller_frame, call_node

    def before_expr(self, node, frame):
        pass

    def after_expr(self, node, frame, value):
        pass

    def before_stmt(self, node, frame):
        pass

    def after_stmt(self, node, frame, exc_value, exc_traceback):
        pass

    def enter_call(self, enter_info):
        pass

    def exit_call(self, exit_info):
        pass

    def parse_extra(self, root, source, filename):
        pass


class _NodeVisitor(ast.NodeTransformer):
    def generic_visit(self, node):
        if isinstance(node, ast.expr) and not (hasattr(node, "ctx") and not isinstance(node.ctx, ast.Load)):
            return self.visit_expr(node)
        if isinstance(node, ast.stmt) and not (isinstance(node, ast.ImportFrom) and node.module == "__future__"):
            return self.visit_stmt(node)
        return super().generic_visit(node)

    def visit_expr(self, node):
        """
        each expression e gets wrapped like this:
            _after(_before(_tree_index), e)
        where
            _after is function that gives the resulting value
            _before is function that signals the beginning of evaluation of e
        """

        if isinstance(node, ast.Starred):
            return super().generic_visit(node)

        before_marker = _create_simple_marker_call(node, TreeTracerBase._treetrace_hidden_before_expr)
        ast.copy_location(before_marker, node)

        after_marker = ast.Call(
            func=ast.Name(id=TreeTracerBase._treetrace_hidden_after_expr.__name__,
                          ctx=ast.Load()),
            args=[
                before_marker,
                super().generic_visit(node),
            ],
            keywords=[],
        )
        ast.copy_location(after_marker, node)
        ast.fix_missing_locations(after_marker)

        return after_marker

    def visit_stmt(self, node):
        wrapped = ast.With(
            items=[ast.withitem(
                context_expr=_create_simple_marker_call(super().generic_visit(node),
                                                        TreeTracerBase._treetrace_hidden_with_stmt))],
            body=[node],
        )
        ast.copy_location(wrapped, node)
        ast.fix_missing_locations(wrapped)
        return wrapped


class _StmtContext:
    __slots__ = ('tracer', 'node', 'frame')

    def __init__(self, tracer, node, frame):
        self.tracer = tracer
        self.node = node
        self.frame = frame

    def __enter__(self):
        tracer = self.tracer
        node = self.node
        frame = self.frame
        if isinstance(node.parent, (ast.FunctionDef, ast.Module)) and node is node.parent.body[0]:
            tracer._enter_call(node, frame)
        frame_info = tracer.stack[frame]
        frame_info.expression_stack = []
        frame_info.statement = node
        tracer.before_stmt(node, frame)

    def __exit__(self, exc_type, exc_val, exc_tb):
        node = self.node
        tracer = self.tracer
        frame = self.frame
        result = tracer.after_stmt(node, frame, exc_val, exc_tb)
        frame_info = tracer.stack[frame]
        if isinstance(node, ast.Return):
            frame_info.return_node = node
        parent = node.parent
        return_node = frame_info.return_node
        exiting = (isinstance(parent, (ast.FunctionDef, ast.Module)) and
                   (node is parent.body[-1] or
                    exc_val or
                    return_node))
        if exiting:
            caller_frame, call_node = tracer._get_caller_stuff(frame)
            return_value = None
            if return_node and return_node.value and not exc_val:
                return_value = frame_info.expression_values[return_node.value]
            tracer.exit_call(ExitCallInfo(call_node,
                                          return_node,
                                          caller_frame,
                                          frame,
                                          return_value,
                                          exc_val,
                                          exc_tb
                                          ))

            tracer.stack.pop(frame)
        return result


def _create_simple_marker_call(node, func):
    """
    :type func: FunctionType
    """
    return ast.Call(
        func=ast.Name(id=func.__name__,
                      ctx=ast.Load()),
        args=[ast.Num(node._tree_index)],
        keywords=[],
    )


def ancestors(node):
    while True:
        try:
            node = node.parent
        except AttributeError:
            break
        yield node


def loops(node):
    result = []
    while True:
        try:
            parent = node.parent
        except AttributeError:
            break
        if isinstance(parent, (ast.FunctionDef, ast.ClassDef)):
            break
        is_containing_loop = (isinstance(parent, ast.For) and parent.iter is not node or
                              isinstance(parent, ast.While) and parent.test is not node)
        if is_containing_loop:
            result.append(parent)
        node = parent
    result.reverse()
    return tuple(result)



from setuptools import setup

setup(name='birdseye',
      version='0.1.1',
      description='Python debugger using the AST',
      classifiers=[
          'Programming Language :: Python :: 3.5',
      ],
      url='http://github.com/alexmojaki/birdseye',
      author='Alex Hall',
      author_email='alex.mojaki@gmail.com',
      license='MIT',
      packages=['birdseye'],
      install_requires=[
          'flask-humanize',
          'flask_sqlalchemy',
          'asttokens',
          'littleutils',
          'qualname',
      ],
      entry_points={
          'console_scripts': ['birdseye=birdseye.server:main'],
      },
      package_data={'': ['static/**/*', 'templates/*']},
      zip_safe=False)



import os

from flask.templating import render_template

from birdseye.app import app, Call, Function, db

from werkzeug.routing import PathConverter

from birdseye.utils import path_leaf, all_file_paths, short_path
from littleutils import strip_required_prefix


class FileConverter(PathConverter):
    regex = '.*?'


app.url_map.converters['file'] = FileConverter


@app.route('/')
def index():
    files = sorted(all_file_paths())
    prefix = os.path.commonprefix(files)
    files = zip(files, [strip_required_prefix(f, prefix) or path_leaf(f)
                        for f in files])
    return render_template('index.html',
                           files=files)


@app.route('/file/<file:path>')
def file_view(path):
    return render_template('file.html',
                           funcs=sorted(db.session.query(Function.name).filter_by(file=path).distinct()),
                           full_path=path,
                           short_path=short_path(path))


@app.route('/file/<file:path>/function/<func_name>')
def func_view(path, func_name):
    query = (db.session.query(Call, Function)
             .join(Function)
             .filter_by(file=path, name=func_name)
             .order_by(Call.start_time.desc())
             [:200])

    func = query[0][1]
    return render_template('function.html',
                           func=func,
                           query=query)


@app.route('/call/<int:call_id>')
def call_view(call_id):
    call = Call.query.filter_by(id=call_id).one()
    func = call.function
    return render_template('call.html',
                           call=call,
                           func=func)


def main():
    app.run(debug=True, port=7777)


if __name__ == '__main__':
    main()
